import pytest
from mock import patch, MagicMock, AsyncMock


@pytest.fixture(scope='function')
def conn():
    p = patch('frankfurt.database.asyncpg.create_pool', new=AsyncMock())
    m = p.start()
    m.return_value.acquire = AsyncMock(return_value = MagicMock())
    conn = m.return_value.acquire.return_value
    conn.__aenter__.return_value = conn
    tr = AsyncMock()
    tr.start = AsyncMock()
    conn.transaction = MagicMock(return_value=tr)
    conn.execute = AsyncMock()
    conn.fetchrow = AsyncMock()
    conn.fetch = AsyncMock()
    yield conn
    p.stop()


@pytest.fixture(scope='function')
def asyncpg_conn():
    p = patch('frankfurt.database.asyncpg.create_pool', new=AsyncMock())
    m = p.start()
    m.return_value.acquire = AsyncMock(return_value = MagicMock())
    conn = m.return_value.acquire.return_value
    conn.__aenter__.return_value = conn
    tr = AsyncMock()
    tr.start = AsyncMock()
    conn.transaction = MagicMock(return_value=tr)
    conn.execute = AsyncMock()
    conn.fetchrow = AsyncMock()
    conn.fetch = AsyncMock()
    yield conn
    p.stop()



@pytest.fixture(scope='function')
def create_pool():
    p = patch('frankfurt.database.asyncpg.create_pool', new=AsyncMock())
    m = p.start()
    m.return_value.acquire = AsyncMock(return_value = MagicMock())
    conn = m.return_value.acquire.return_value
    conn.__aenter__.return_value = conn
    tr = AsyncMock()
    tr.start = AsyncMock()
    conn.transaction = MagicMock(return_value=tr)
    conn.execute = AsyncMock()
    conn.fetchrow = AsyncMock()
    conn.fetch = AsyncMock()
    yield m
    p.stop()



import pytest


def test_types():
    from frankfurt import fields

    f = fields.IntegerField()
    f.assert_type(1)
    f.assert_type(0)
    f.assert_type(-1)
    f.assert_type(None)

    f = fields.IntegerField(not_null=True)
    f.assert_type(1)
    f.assert_type(0)
    f.assert_type(-1)
    with pytest.raises(TypeError):
        f.assert_type(None)

    f = fields.IntegerField(non_negative=True)
    f.assert_type(1)
    f.assert_type(0)
    with pytest.raises(TypeError):
        f.assert_type(-1)
    f.assert_type(None)

    f = fields.IntegerField(non_negative=True)
    f.assert_type(1)
    f.assert_type(0)
    with pytest.raises(TypeError):
        f.assert_type(-1)
    f.assert_type(None)


def test_general_arguments():
    from frankfurt import fields

    with pytest.raises(TypeError):
        fields.UUIDField(pk=True)


@pytest.mark.asyncio
async def test_fields_by_examples(asyncpg_conn):

    from frankfurt import Database, fields
    from frankfurt.models import Model

    class Example_1(Model):
        password = fields.BinaryField()

    db = await Database('__url__', models=[Example_1])

    async with db.acquire() as conn:
        await conn.create_all_tables()

    asyncpg_conn.execute.assert_awaited_once_with(
        'CREATE TABLE IF NOT EXISTS "example_1" ("password" BYTEA)'
    )


def test_default_values():

    from frankfurt import fields
    from frankfurt.models import Model

    class Example(Model):
        text_1 = fields.CharField(max_length=200, default='text')
        text_2 = fields.CharField(max_length=200, default=None)
        text_3 = fields.CharField(max_length=200, default=lambda : "sentinel")

    example = Example()
    assert example['text_1'] == 'text'
    assert example['text_2'] is None
    assert example['text_3'] == 'sentinel'


def test_not_null_values():

    from frankfurt import fields
    from frankfurt.models import Model

    class Example(Model):
        text = fields.CharField(max_length=200, not_null=True)

    example = Example()

    with pytest.raises(KeyError):
        assert example['text'] is None


@pytest.mark.asyncio
async def test_unique_constraint(asyncpg_conn):

    from frankfurt import Database, fields
    from frankfurt.models import Model

    class Example(Model):
        token = fields.CharField(max_length=64, unique=True)

    db = await Database('__url__', models=[Example])

    async with db.acquire() as conn:
        await conn.create_all_tables()

    asyncpg_conn.execute.assert_awaited_once_with(
        'CREATE TABLE IF NOT EXISTS "example" ("token" VARCHAR(64) UNIQUE)'
    )



import dataclasses

from typing import Tuple, Type, Dict, Any, List, Optional

from .fields import BaseField, ForeignKeyField


@dataclasses.dataclass
class Meta:
    name: str

    abstract : bool = False

    table_name : str = dataclasses.field(init=False)

    fields : Dict[str, BaseField] = dataclasses.field(
        default_factory=dict
    )

    unique_fields : List[str] = dataclasses.field(
        default_factory=list
    )

    db: Optional = None

    primary_key : List[str] = dataclasses.field(
        default_factory=list
    )

    polymorphic_bases : List[str] = dataclasses.field(
        default_factory=list
    )

    polymorphic_on : Optional[str] = None

    polymorphic_identity : Optional[Any] = None

    def __post_init__(self):
        self.table_name = self.name.lower()

    @property
    def depends(self):
        # Return a list of methods that should be created first.

        deps = []

        for field_ in self.fields.values():
            if isinstance(field_, ForeignKeyField):
                deps.append(field_.to_field.model)

        return deps

    def extend_from_model(self, Meta):

        # Set the db.
        if hasattr(Meta, 'db'):
            self.db = Meta.db

        # Change the abstract.
        self.abstract = getattr(Meta, "abstract", self.abstract)

        # Change the table name.
        if hasattr(Meta, 'table_name'):
            self.table_name = Meta.table_name

        # Update the polymorphic information
        if hasattr(Meta, 'polymorphic_identity'):
            self.polymorphic_identity = Meta.polymorphic_identity

        if hasattr(Meta, 'polymorphic_on'):
            self.polymorphic_on = Meta.polymorphic_on


class ModelType(type):

    # Properties in the type
    @property
    def table_name(cls):
        return cls._meta.table_name

    def __new__(cls, name: str, bases: Tuple[Type, ...], attrs: Dict[str, Any]):

        # Start with a fresh meta.
        meta = Meta(name)

        # Extract some information from bases.
        for base in bases:
            if hasattr(base, '_meta'):

                # Copy the reference to a database.
                meta.db = base._meta.db

                # Copy fields.
                for field_name, field in base._meta.fields.items():
                    meta.fields[field_name] = field.copy()

                    # Check if field is unique.
                    if field.unique:
                        meta.unique_fields.append(field_name)

                # Extract the polymorphic_on property.
                meta.polymorphic_on = base._meta.polymorphic_on
                meta.polymorphic_bases = base._meta.polymorphic_bases[:]

        if "Meta" in attrs:
            meta.extend_from_model(attrs["Meta"])

        # Extend the polymorphic bases
        if meta.polymorphic_on is not None:
            meta.polymorphic_bases.append(meta.table_name)

        # Append the fields defined in the model.
        for fname, value in attrs.items():
            if isinstance(value, BaseField):
                meta.fields[fname] = value

        # Detect the primary key from the fields.
        for fname, field in meta.fields.items():
            if field.primary_key:
                meta.primary_key.append(fname)

        # Modify the primary keys in case of a polymorphic base.
        if meta.polymorphic_on in meta.fields:
            if meta.table_name != meta.polymorphic_bases[0]:
                for fname in meta.primary_key:
                    meta.fields[fname] = ForeignKeyField(
                        to=f'{meta.polymorphic_bases[0]}.{fname}',
                        primary_key=True
                    )

        # Append the meta information to the class.
        attrs["_meta"] = meta

        model = super().__new__(cls, name, bases, attrs)

        # Add a reference for the fields.
        for field in meta.fields.values():
            field.model = model

        return model

    def from_record(cls, record):
        return cls(**record)


class Model(metaclass=ModelType):
    def __init__(self, **kwargs):

        # Save the value of the fields in this dict
        self._data_default = {}
        self._data = {}

        # Check for the kwargs in the fields.
        for name, value in kwargs.items():
            self._set_data(name, value)

        # Continue with default values.
        for name, field in self._meta.fields.items():
            if name not in self._data:
                if field.has_default:
                    self._set_data_default(name, field.default)

        # Check for not nulls (or maybe types?)
        for name, field in self._meta.fields.items():
            msg = f"Field {name} cannot be null (None)"

            if name in self._data:
                if field.not_null and self._data[name] is None:
                    raise TypeError(msg)
            elif name in self._data_default:
                if field.not_null and self._data_default[name] is None:
                    raise TypeError(msg)

    def __repr__(self):
        return f'<{self.__class__.__name__} />'

    def _set_data(self, name, value):
        if name not in self._meta.fields:
            raise TypeError(f"Model has no field {name}.")

        # Assert the type for this field.
        try:
            self._meta.fields[name].assert_type(value)
            self._data[name] = value
        except TypeError as e:
            msg = f"Field {name} can't accept value {value}: {str(e).lower()}"
            raise TypeError(msg)

    def _set_data_default(self, name, value):
        try:
            self._meta.fields[name].assert_type(value)
            self._data_default[name] = value
        except TypeError as e:
            msg = f"Field {name} can't accept value {value}: {str(e).lower()}"
            raise TypeError(msg)

    def __contains__(self, key):
        return (key in self._data) or (key in self._data_default)

    def __setitem__(self, key : str, value):
        if key not in self._meta.fields:
            msg = "Model {} has not field {}.".format(self._meta.name, key)
            raise KeyError(msg)

        self._data[key] = value

        if key in self._data_default:
            del self._data_default[key]

    def __getitem__(self, key : str):
        if key not in self._meta.fields:
            msg = "Model {} has not field {}.".format(self._meta.name, key)
            raise KeyError(msg)

        if key in self._data:
            return self._data[key]

        if key in self._data_default:
            return self._data_default[key]

        raise KeyError("Field {} has no value.".format(key))

    def _update_data(self, **kwargs):
        # Update the inner data from a record.

        if not set(kwargs).issubset(self._meta.fields):
            raise TypeError("Some arguments are not defined in the model.", kwargs)

        for name, value in kwargs.items():

            if name in self._meta.fields:
                self._data[name] = value

                if name in self._data_default:
                    del self._data_default[name]

        return self


class DeleteMixin:
    async def delete(self, sess):
        await sess.delete(self)


class UpdateMixin:
    async def update(self, sess):
        # Update a row in the database.
        return await sess.update(self)


class CreateMixin:
    @classmethod
    async def create(cls, db, **kwargs):
        # Create a row in the database.
        return await db.insert(cls(**kwargs))


class InsertMixin:
    async def insert(self, sess):
        # Insert a row in the database.
        return await sess.insert(self)



import pytest
from mock import AsyncMock


@pytest.mark.asyncio
async def test_create_table_no_columns():
    import frankfurt.sql

    conn = AsyncMock()

    await frankfurt.sql.PartialCreateTable(
        table_name='test'
    ).execute(conn=conn)

    conn.execute.assert_called_once_with(
        'CREATE TABLE IF NOT EXISTS "test"'
    )


@pytest.mark.asyncio
async def test_create_table_with_many_columns():
    import frankfurt.sql

    conn = AsyncMock()

    columns = {
        'col_0': frankfurt.sql.ColumnClause('dt_0'),
        'col_1': frankfurt.sql.ColumnClause('dt_1')
    }

    ct = frankfurt.sql.PartialCreateTable(table_name='test')
    ct.columns(**columns)
    await ct.execute(conn=conn)

    conn.execute.assert_called_once_with(
        'CREATE TABLE IF NOT EXISTS "test" ("col_0" dt_0, "col_1" dt_1)'
    )


@pytest.mark.asyncio
async def test_simple_insert():
    import frankfurt.sql

    conn = AsyncMock()
    conn.fetchrow.return_value = {
        'row': (1, 2)
    }

    ins = frankfurt.sql.PartialInsert(table_name='test')
    ins.values(a=1)
    ins.returning('a', 'b')

    record = await ins.execute(conn=conn)

    assert record['a'] == 1
    assert record['b'] == 2

    conn.fetchrow.assert_called_with(
        'INSERT INTO "test" ("a") VALUES ($1) RETURNING ("a", "b")', 1
    )


@pytest.mark.asyncio
async def test_delete_full_table():
    import frankfurt.sql

    conn = AsyncMock()

    # Define.
    delete = frankfurt.sql.PartialDelete(table_name="table")

    # Execute on await.
    await delete.execute(conn=conn)

    conn.fetch.assert_awaited_once_with(
        'DELETE FROM "table"'
    )


@pytest.mark.asyncio
async def test_delete_where():
    import frankfurt.sql

    conn = AsyncMock()

    # Define.
    delete = frankfurt.sql.PartialDelete(table_name="table")
    delete.where(pk=1, text="hola")

    # Execute on await.
    await delete.execute(conn=conn)

    # Assert for where values.
    conn.fetch.assert_awaited_once_with(
        'DELETE FROM "table" WHERE "pk"=$1 AND "text"=$2', 1, 'hola'
    )



import asyncpg

from typing import Type, Dict, Optional

from .models import Model
from .fields import ForeignKeyField, Action

from . import sql


class Database:
    """
    Database
    ========

    Database instance to interact with asyncpg.

    """

    #: Keep track of all the models defined in this database.
    models: Dict[str, Type[Model]]

    debug : bool = False

    _pool : Optional[asyncpg.pool.Pool] = None

    def __init__(self, dsn : str, models=None, debug=False, pool_kwargs=None):
        # Initialize a Database instance.

        if models is None:
            models = []

        # Store the information for the pool
        self._dsn = dsn
        self._pool_kwargs = pool_kwargs if pool_kwargs is not None else {}

        # Extract the models mapped by table_name.
        self.models = {}
        for model in models:
            model._meta.db = self
            self.models[model.table_name] = model

    def __await__(self):

        # Create the pool.
        self._pool = yield from asyncpg.create_pool(
            dsn=self._dsn, **self._pool_kwargs
        ).__await__()

        return self

    def acquire(self):
        return Connection(db=self)

    async def release(self, sess):
        await self._pool.release(sess.conn)

    # async def save(self, model, conn=None):

    #     if conn is None:
    #         conn = self._conn.get()

    #     # Update with values.
    #     values = {}

    #     # Find a where clause.
    #     where_clause = sql.WhereClause()

    #     # Find our pk.
    #     for name, field in model._meta.fields.items():
    #         if name in model._data and name in model._meta.primary_key:
    #             where_clause &= {name: model._data[name]}
    #         elif name in model._data:
    #             values[name] = model._data[name]
    #         elif name in model._data_default:
    #             values[name] = model._data_default[name]

    #     if where_clause:
    #         # Create an update.
    #         stmt = sql.PartialUpdate(
    #             conn, model._meta.table_name
    #         ).where(
    #             where_clause
    #         )
    #     else:
    #         # Insert statement
    #         stmt = sql.PartialInsert(
    #             conn, model._meta.table_name
    #         )

    #     # Update the rest.
    #     stmt.values(**values)

    #     stmt.returning(*model._meta.fields.keys())

    #     return model._update_data(**await stmt)


class Connection:
    def __init__(self, db):
        self.db = db

    def __await__(self):

        # Start a connection.
        self._conn = yield from self.db._pool.acquire().__await__()

        # Create a transaction.
        self._tr = self._conn.transaction()

        # Start the transaction.
        yield from self._tr.start().__await__()

        return self

    async def __aenter__(self):
        return await self

    async def __aexit__(self, exc_type, exc_value, traceback):
        if exc_type is None:
            await self._tr.commit()
        else:
            await self._tr.rollback()

        # Release the connection.
        await self.db._pool.release(self._conn)

    async def commit(self):
        await self._tr.commit()
        self._tr = self._conn.transaction()
        await self._tr.start()

    async def rollback(self):
        await self._tr.commit()
        self._tr = self._conn.transaction()
        await self._tr.start()

    async def insert(self, instance):
        # Insert an instance into the database.

        stmt = sql.PartialInsert(
            table_name=instance._meta.table_name
        )

        # Find our pk.
        for name, field in instance._meta.fields.items():
            if name in instance._data:
                stmt.set_value(name, instance._data[name])
            elif name in instance._data_default:
                stmt.set_value(name, instance._data_default[name])

        # Return all the columns.
        stmt.returning(*instance._meta.fields.keys())
        stmt.with_conn(self._conn)

        # Update the instance.
        record = await stmt

        return instance._update_data(**record)

    def delete(self, instance):

        # Create the delete statement.
        stmt = sql.PartialDelete(
            table_name=instance._meta.table_name,
            mapper=lambda x: instance.__class__(**x)
        )

        # Find the where clause.
        for name in instance._meta.primary_key:
            if name in instance._data:
                stmt.where(**{name: instance._data[name]})

        if not stmt._where:
            raise Exception("Instance has no primary key.")

        stmt.returning(*instance._meta.fields.keys())

        return stmt.with_conn(self._conn)

    def select(self, model):

        stmt = sql.PartialSelect(
            table_name=model._meta.table_name,
            mapper=lambda _: model(**_)
        )

        stmt.select(*list(model._meta.fields.keys()))
        stmt.with_conn(self._conn)

        return stmt

    async def update(self, instance):

        # Create an update.
        stmt = sql.PartialUpdate(
            table_name=instance._meta.table_name
        )

        # Find a where clause.
        where_clause = sql.WhereClause()

        # Find our pk.
        for name, field in instance._meta.fields.items():
            if name in instance._data and name in instance._meta.primary_key:
                where_clause &= {name: instance._data[name]}
            elif name in instance._data:
                stmt.set_value(name, instance._data[name])
            elif name in instance._data_default:
                stmt.set_value(name, instance._data_default[name])

        if not where_clause:
            raise Exception("Unable to determine a where clause.")

        # Add the where clause and the returning columns.
        stmt.where(where_clause)
        stmt.returning(*instance._meta.fields.keys())

        # Update data.
        record = await stmt.execute(conn=self._conn)

        return instance._update_data(**record)

    async def create_all_tables(self):

        # Sort the models in the right order.
        models_unsorted = list(self.db.models.keys())
        models_sorted = []

        while len(models_unsorted) > 0:
            key = models_unsorted[0]
            deps = self.db.models[key]._meta.depends
            if all(_._meta.table_name in models_sorted for _ in deps):
                models_sorted.append(models_unsorted.pop(0))
            else:
                models_unsorted.append(models_unsorted.pop(0))

        # Run all the statements inside a transaction.
        for table_name in models_sorted:
            await self.create_table(table_name)

    async def create_table(self, table_name : str):

        # Get the model.
        model = self.db.models[table_name]

        # Create table statement.
        stmt = sql.PartialCreateTable(
            table_name=table_name
        )

        for field_name, field in model._meta.fields.items():

            # Resolve the constraint.
            constraint = sql.ColumnConstraint(
                not_null=field.not_null,
                unique=field.unique,
            )

            if len(model._meta.primary_key) < 2:
                constraint.primary_key = field.primary_key

            # Resolve any dependecies.
            if isinstance(field, ForeignKeyField):
                reftable, refcolumn = field.refs()

                # Get the on_delete action.
                on_delete = None
                if isinstance(field.on_delete, Action):
                    on_delete = field.on_delete.value

                references = sql.ColumnConstraintReference(
                    reftable, refcolumn=refcolumn,
                    on_delete=on_delete
                )

                constraint.references = references

            # Append a new column.
            stmt.columns(**{
                field_name: sql.ColumnClause(
                    field.__postgresql_type__(),
                    constraint
                )
            })

        # Append the table constraints
        if len(model._meta.primary_key) > 1:
            stmt.constraints(sql.TableConstraintPrimaryKey(
                keys=model._meta.primary_key
            ))

        # Attach a connection.
        stmt.with_conn(self._conn)

        await stmt



import pytest


@pytest.mark.asyncio
async def test_create_pool(create_pool):
    from frankfurt import Database, fields
    from frankfurt.models import Model

    class Example(Model):
        text = fields.CharField(max_length=200)

    await Database('__url__', models=[Example])

    # Assert that asyncpg is called.
    create_pool.assert_awaited_once_with(dsn='__url__')


@pytest.mark.asyncio
async def test_conn(create_pool):
    from frankfurt import Database, fields
    from frankfurt.models import Model

    class Example(Model):
        text = fields.CharField(max_length=200)

    db = await Database('__url__', models=[Example])

    async with db.acquire() as conn:
        await conn.create_all_tables()

    conn._conn.execute.assert_called_once_with(
        'CREATE TABLE IF NOT EXISTS "example" ("text" VARCHAR(200))'
    )

    # The connection must be acquired and released.
    db._pool.acquire.assert_awaited_once_with()
    db._pool.release.assert_awaited_once_with(conn._conn)

    # The transaction should be commited.
    conn._tr.commit.assert_awaited_once_with()



import pytest
from mock import call


@pytest.mark.asyncio
async def test_simple_inheritance(asyncpg_conn):

    from frankfurt import Database, fields
    from frankfurt.models import Model

    class BaseModel(Model):
        pk = fields.UUIDField(primary_key=True)

        class Meta:
            abstract = True

    class Example1(BaseModel):
        text1 = fields.CharField(max_length=200)

    class Example2(BaseModel):
        text2 = fields.CharField(max_length=200)

    db = await Database('__url__', models=[Example1, Example2])

    async with db.acquire() as conn:
        await conn.create_all_tables()

    assert asyncpg_conn.execute.await_args_list == [
        call('CREATE TABLE IF NOT EXISTS "example1" '
             '("pk" UUID PRIMARY KEY, "text1" VARCHAR(200))'),
        call('CREATE TABLE IF NOT EXISTS "example2" '
             '("pk" UUID PRIMARY KEY, "text2" VARCHAR(200))'),
    ]


@pytest.mark.asyncio
async def test_polymorphic_models(asyncpg_conn):

    from frankfurt import Database, fields
    from frankfurt.models import Model

    class BaseModel(Model):
        pk = fields.UUIDField(primary_key=True)

        class Meta:
            abstract = True

    class Figure(BaseModel):

        shape = fields.CharField(max_length=200, not_null=True)

        class Meta:
            polymorphic_on = 'shape'
            polymorphic_identity = 'figure'

    class Rect(Figure):

        height = fields.IntegerField(
            default=0, non_negative=True, not_null=True
        )

        width = fields.IntegerField(
            default=0, non_negative=True, not_null=True
        )

        class Meta:
            polymorphic_identity = 'rect'

    class Circle(Figure):

        radius = fields.IntegerField(
            default=0, non_negative=True, not_null=True
        )

        class Meta:
            polymorphic_identity = 'rect'

    db = await Database('__url__', models=[Rect, Circle, Figure])

    async with db.acquire() as conn:
        await conn.create_all_tables()

    assert asyncpg_conn.execute.await_args_list == [
        call('CREATE TABLE IF NOT EXISTS "figure" '
             '("pk" UUID PRIMARY KEY, "shape" VARCHAR(200) NOT NULL)'),
        call('CREATE TABLE IF NOT EXISTS "rect" ('
             '"pk" UUID PRIMARY KEY REFERENCES "figure" ("pk"), '
             '"shape" VARCHAR(200) NOT NULL, '
             '"height" INTEGER NOT NULL, "width" INTEGER NOT NULL)'),
        call('CREATE TABLE IF NOT EXISTS "circle" ('
             '"pk" UUID PRIMARY KEY REFERENCES "figure" ("pk"), '
             '"shape" VARCHAR(200) NOT NULL, '
             '"radius" INTEGER NOT NULL)'),
    ]



import logging
import dataclasses
import asyncpg

from typing import Dict, Optional, List, Any, Union


logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


@dataclasses.dataclass
class StatementContext:
    values : List[Any] = dataclasses.field(default_factory=list)


class Statement:
    def __stmt__(self, ctx : StatementContext):
        # Statements are built using the magic method __stmt__
        return str(self)


@dataclasses.dataclass
class ColumnConstraintReference:
    reftable: str
    refcolumn: Optional[str] = None
    on_delete: Optional[str] = None
    on_update: Optional[str] = None

    def __str__(self):
        stmt = f'REFERENCES "{self.reftable}"'

        if self.refcolumn:
            stmt = f'{stmt} ("{self.refcolumn}")'

        if self.on_delete:
            stmt = f'{stmt} ON DELETE {self.on_delete}'

        return stmt


@dataclasses.dataclass
class TableConstraint(Statement):
    pass


@dataclasses.dataclass
class TableConstraintPrimaryKey(TableConstraint):
    keys : List[str] = dataclasses.field(default_factory=list)

    def __bool__(self):
        return len(self.keys) > 0

    def __stmt__(self, ctx):
        keys = ', '.join(f'"{k}"' for k in self.keys)
        return f'PRIMARY KEY ({keys})'


@dataclasses.dataclass
class ColumnConstraint:
    not_null: bool = False
    primary_key: bool = False
    references: Optional[ColumnConstraintReference] = None
    unique : bool = False

    def __bool__(self):

        if self.not_null:
            return True

        if self.primary_key:
            return True

        if self.references:
            return True

        if self.unique:
            return True

        return False

    def __str__(self):

        constraints = []

        if self.not_null:
            constraints.append("NOT NULL")

        if self.unique:
            constraints.append("UNIQUE")

        if self.primary_key:
            constraints.append("PRIMARY KEY")

        if self.references:
            constraints.append(str(self.references))

        return " ".join(constraints)


@dataclasses.dataclass
class ColumnClause(Statement):
    data_type : str
    constraint : Optional[ColumnConstraint] = None

    def __stmt__(self, ctx):
        stmt = f'{self.data_type}'

        if self.constraint:
            stmt = f"{stmt} {self.constraint}"
        return stmt


@dataclasses.dataclass
class WhereClause(Statement):
    def __bool__(self):
        return False

    def __and__(self, other):

        if isinstance(other, dict):
            parts = []
            for n, v in other.items():
                lhs = ExpressionColumnName(n)
                rhs = ExpressionLiteralValue(v)
                parts.append(WhereClauseEquality(lhs=lhs, rhs=rhs))
            return WhereClauseAnd(parts=parts)

        if isinstance(other, WhereClauseAnd):
            return WhereClauseAnd(parts=other.parts)

        if isinstance(other, WhereClause):
            return WhereClauseAnd(parts=[other])

        raise NotImplementedError("Operation not implemented")


@dataclasses.dataclass
class SetClause(Statement):
    values : Dict[str, Any] = dataclasses.field(default_factory=dict)

    def __bool__(self):
        return len(self.values) > 0

    def __stmt__(self, ctx):
        set_values = []
        for name, value in self.values.items():
            ctx.values.append(value)
            set_values.append(f'"{name}"=${len(ctx.values)}')

        return ', '.join(set_values)


@dataclasses.dataclass
class ReturningClause(Statement):
    names : List[str] = dataclasses.field(default_factory=list)

    def __bool__(self):
        return len(self.names) > 0

    def __stmt__(self, ctx):
        return ', '.join(f'"{name}"' for name in self.names)

    def __call__(self, *args):
        self.names.extend(args)


@dataclasses.dataclass
class SelectClause(Statement):
    names : List[str] = dataclasses.field(default_factory=list)

    def __bool__(self):
        return len(self.names) > 0

    def __stmt__(self, ctx):
        return ', '.join(f'"{name}"' for name in self.names)


@dataclasses.dataclass
class ValuesClause(Statement):
    values : Dict[str, Any] = dataclasses.field(default_factory=dict)

    def __bool__(self):
        return len(self.values) > 0

    def __stmt__(self, ctx):

        cols = []
        placeholders = []

        for name, value in self.values.items():
            ctx.values.append(value)
            cols.append(f'"{name}"')
            placeholders.append(f'${len(ctx.values)}')

        return f'({", ".join(cols)}) VALUES ({", ".join(placeholders)})'


class Expression(Statement):
    def __bool__(self):
        return False


@dataclasses.dataclass
class WhereClauseEquality(WhereClause):
    lhs: Union[Expression, str]
    rhs: Union[Expression, str]

    def __bool__(self):
        return True

    def __stmt__(self, ctx):
        return f'{self.lhs.__stmt__(ctx)}={self.rhs.__stmt__(ctx)}'


@dataclasses.dataclass
class WhereClauseAnd(WhereClause):
    parts: List[WhereClause] = dataclasses.field(default_factory=list)

    def __bool__(self):
        return len(self.parts) > 0

    def __str__(self):
        return " AND ".join(f'({_})' for _ in self.parts)

    def __stmt__(self, ctx):
        return " AND ".join(f'{_.__stmt__(ctx)}' for _ in self.parts)

    def __and__(self, other):
        # `and` operation between WhereClause's.

        if isinstance(other, WhereClauseAnd):
            return WhereClauseAnd(parts=self.parts + other.parts)

        if isinstance(other, WhereClause):
            return WhereClauseAnd(parts=self.parts + [other])

        if isinstance(other, dict):
            return self & (WhereClause() & other)

        raise NotImplementedError("Operation not implemented")


@dataclasses.dataclass
class ExpressionColumnName(Expression):
    name : str

    def __bool__(self):
        return self.name != ""

    def __str__(self):
        return f'"{self.name}"'


@dataclasses.dataclass
class ExpressionLiteralValue(Expression):
    value : Any

    def __bool__(self):
        return str(self.value) != ""

    def __stmt__(self, ctx):

        ctx.values.append(self.value)

        i = len(ctx.values)

        return f'${i}'


class Asterisk(Expression):
    def __bool__(self):
        return True

    def __str__(self):
        return '*'


@dataclasses.dataclass
class WhereMixin:

    _where : WhereClause = dataclasses.field(
        init=False, default_factory=WhereClause
    )

    def where(self, *args, **kwargs):

        # When args are passed are considered to be where clauses.
        for arg in args:
            self._where &= arg

        for name, value in kwargs.items():
            lhs = ExpressionColumnName(name=name)
            rhs = ExpressionLiteralValue(value=value)
            self._where &= WhereClauseEquality(lhs=lhs, rhs=rhs)

        return self


@dataclasses.dataclass
class ReturningMixin:

    _returning : ReturningClause = dataclasses.field(
        init=False, default_factory=ReturningClause
    )

    def returning(self, *names):
        self._returning.names.extend(names)


@dataclasses.dataclass
class SelectMixin:

    _select : SelectClause = dataclasses.field(
        init=False, default_factory=SelectClause
    )

    def select(self, *names):
        self._select.names.extend(names)
        return self


@dataclasses.dataclass
class ValuesMixin:
    _values : ValuesClause = dataclasses.field(
        init=False, default_factory=ValuesClause
    )

    def values(self, **kwargs):
        self._values.values.update(kwargs)
        return self

    def set_value(self, key, value):
        self._values.values[key] = value


@dataclasses.dataclass
class SetMixin:
    _set : SetClause = dataclasses.field(
        init=False, default_factory=SetClause
    )

    def values(self, **kwargs):
        self._set.values.update(kwargs)
        return self

    def set_value(self, key, value):
        self._set.values[key] = value


@dataclasses.dataclass
class Command:
    conn : asyncpg.Connection = dataclasses.field(
        init=False
    )

    async def execute(self, conn=None):
        if conn is not None:
            self.conn = conn
        return await self

    def with_conn(self, conn):
        self.conn = conn
        return self


@dataclasses.dataclass
class PartialDelete(Command, WhereMixin, ReturningMixin):

    table_name : str

    mapper : Optional[callable] = None

    def __await__(self):

        # Start with a statement context.
        ctx = StatementContext()

        stmt = f'DELETE FROM "{self.table_name}"'

        # Append a where clause (Mmm, en error here could
        # delete a whole table.)
        if self._where:
            stmt = f'{stmt} WHERE {self._where.__stmt__(ctx)}'

        # Append the returning columns.
        if self._returning:
            stmt = f"{stmt} RETURNING ({self._returning.__stmt__(ctx)})"

        logger.debug(stmt)
        record = yield from self.conn.fetch(stmt, *ctx.values).__await__()

        if self._returning:
            return {k: v for k, v in zip(self._returning.names, record['row'])}

        return None


@dataclasses.dataclass
class PartialUpdate(Command, WhereMixin, ReturningMixin, SetMixin):

    table_name : str

    mapper : Optional[callable] = None

    def __await__(self):

        # Start with a statement context.
        ctx = StatementContext()

        stmt = f'UPDATE "{self.table_name}"'

        # Append a set clause.
        if self._set:
            stmt = f'{stmt} SET {self._set.__stmt__(ctx)}'

        # Append a where clause.
        if self._where:
            stmt = f'{stmt} WHERE {self._where.__stmt__(ctx)}'

        # Append the returning columns.
        if self._returning:
            stmt = f"{stmt} RETURNING ({self._returning.__stmt__(ctx)})"

        # Wait for the update.
        logger.debug(stmt)
        result = yield from self.conn.fetchrow(stmt, *ctx.values).__await__()

        if callable(self.mapper):
            return self.mapper(result)

        return result


@dataclasses.dataclass
class PartialSelect(Command, WhereMixin, SelectMixin):

    table_name : str

    mapper : Optional[callable] = None

    _one : bool = dataclasses.field(init=False, default=False)

    def one(self, conn=None):
        #: Return one instance of a model or raise an error.
        self._one = True

        return self

    def __await__(self):

        ctx = StatementContext()

        stmt = f'SELECT {self._select.__stmt__(ctx)} FROM "{self.table_name}"'

        if self._where:
            stmt = f"{stmt} WHERE {self._where.__stmt__(ctx)}"

        # Wait for the update.
        logger.debug(stmt)
        records = (yield from self.conn.fetch(stmt, *ctx.values).__await__())

        if callable(self.mapper):
            return self.mapper(records[0])

        return records[0]


@dataclasses.dataclass
class PartialInsert(Command, ReturningMixin, ValuesMixin):
    """
    INSERT Command
    ==============

    Construct an INSERT command and execute it when awaited.
    """

    table_name : str

    mapper : Optional[callable] = None

    def __await__(self):

        ctx = StatementContext()

        stmt = f'INSERT INTO "{self.table_name}"'

        if self._values:
            stmt = f"{stmt} {self._values.__stmt__(ctx)}"

        if self._returning:
            stmt = f"{stmt} RETURNING ({self._returning.__stmt__(ctx)})"

        logger.debug(stmt)
        record = yield from self.conn.fetchrow(stmt, *ctx.values).__await__()

        if self._returning:
            record = {k: v for k, v in zip(self._returning.names, record['row'])}
            if callable(self.mapper):
                return self.mapper(record)
            return record

        return None


@dataclasses.dataclass
class PartialCreateTable(Command):

    table_name : str

    _columns : Dict[str, ColumnClause] = dataclasses.field(
        default_factory=dict
    )

    _constraints : List[TableConstraint] = dataclasses.field(
        default_factory=list
    )

    def columns(self, **kwargs):
        self._columns.update(kwargs)
        return self

    def constraints(self, *args):
        self._constraints.extend(args)
        return self

    def __await__(self):

        ctx = StatementContext()

        stmt = f'CREATE TABLE IF NOT EXISTS "{self.table_name}"'

        # Columns and constraints are mixed.
        columns_and_constraints = []

        for name, col in self._columns.items():
            columns_and_constraints.append(f'"{name}" {col.__stmt__(ctx)}')

        for cons in self._constraints:
            if cons:
                columns_and_constraints.append(f'{cons.__stmt__(ctx)}')

        if columns_and_constraints:
            stmt = f'{stmt} ({", ".join(columns_and_constraints)})'

        logger.debug(stmt)
        yield from self.conn.execute(stmt).__await__()



from uuid import uuid4
import pytest
from mock import call


def test_simple_model():

    from frankfurt.models import Model
    from frankfurt import fields

    class Example(Model):
        text = fields.CharField(max_length=200)

    example = Example(text='test')
    assert example['text'] == 'test'

    with pytest.raises(KeyError):
        example['text_3']


def test_exception_incorrect_argument():
    from frankfurt.models import Model
    from frankfurt import fields

    class Example(Model):
        text = fields.CharField(max_length=200)

    with pytest.raises(TypeError):
        Example(another='test')


def test_abstract_model():

    from frankfurt.models import Model

    class Example(Model):
        class Meta:
            abstract = True

    assert Example._meta.abstract


@pytest.mark.asyncio
async def test_two_primary_keys(asyncpg_conn):
    from frankfurt import Database
    from frankfurt import fields
    from frankfurt.models import Model

    class Book(Model):
        author = fields.CharField(max_length=200, primary_key=True)
        title = fields.CharField(max_length=200, primary_key=True)

    # Get a session.
    db = await Database('__url__', models=[Book])

    async with db.acquire() as conn:
        await conn.create_all_tables()

    # Assert the creation.
    asyncpg_conn.execute.assert_awaited_once_with(
        'CREATE TABLE IF NOT EXISTS "book" ('
        '"author" VARCHAR(200), "title" VARCHAR(200), '
        'PRIMARY KEY ("author", "title"))'
    )


def test_inherit_fields():
    from frankfurt import fields
    from frankfurt.models import Model

    class Writer(Model):
        pk = fields.UUIDField(primary_key=True)
        full_name = fields.CharField(max_length=400)

        class Meta:
            table_name = 'writers'

    class BaseModel(Model):
        pk = fields.UUIDField(primary_key=True)
        creation_date = fields.DateTimeField(auto_now_add=True)
        writer = fields.ForeignKeyField(to='writers')

        class Meta:
            abstract = True

    class Book(BaseModel):
        title = fields.CharField(max_length=200)

    # Check that book has the right fields.
    assert 'pk' in Book._meta.fields
    assert 'creation_date' in Book._meta.fields
    assert 'writer' in Book._meta.fields

    # Check that the fields are new instances.
    assert BaseModel._meta.fields['pk'] != Book._meta.fields['pk']
    assert BaseModel._meta.fields['pk'].model is BaseModel
    assert Book._meta.fields['pk'].model is Book


@pytest.mark.asyncio
async def test_create_table_from_model(asyncpg_conn):
    from frankfurt import Database, fields
    from frankfurt.models import Model

    class Writer(Model):
        pk = fields.UUIDField(primary_key=True)
        full_name = fields.CharField(max_length=400)

        class Meta:
            table_name = 'writers'

    class Book(Model):
        title = fields.CharField(max_length=200)
        date = fields.DateTimeField()
        author = fields.ForeignKeyField(
            to='writers', on_delete=fields.CASCADE
        )

        class Meta:
            table_name = 'books'

    db = await Database('__url__', models=[Writer, Book])

    async with db.acquire() as conn:
        await conn.create_table('books')

    # Assert the connection.
    asyncpg_conn.execute.assert_awaited_once_with(
        f'CREATE TABLE IF NOT EXISTS "books" ("title" VARCHAR(200), '
        '"date" TIMESTAMPTZ, "author" UUID REFERENCES "writers" ("pk") ON '
        'DELETE CASCADE)'
    )


@pytest.mark.asyncio
async def test_create_all_tables_1(asyncpg_conn):

    from frankfurt import Database, fields
    from frankfurt.models import Model

    class Writer(Model):
        pk = fields.UUIDField(primary_key=True)
        full_name = fields.CharField(max_length=400)

        class Meta:
            table_name = 'writers'

    class Book(Model):
        title = fields.CharField(max_length=200)
        date = fields.DateTimeField()
        author = fields.ForeignKeyField(to='writers')

        class Meta:
            table_name = 'books'

    db = await Database('__url__', models=[Writer, Book])

    async with db.acquire() as conn:
        await conn.create_all_tables()

    # Assert the connection.
    assert asyncpg_conn.execute.mock_calls == [
        call('CREATE TABLE IF NOT EXISTS "writers" '
             '("pk" UUID PRIMARY KEY, "full_name" VARCHAR(400))'),
        call(f'CREATE TABLE IF NOT EXISTS "books" ("title" VARCHAR(200), '
             '"date" TIMESTAMPTZ, "author" UUID REFERENCES "writers" ("pk"))'),
    ]


@pytest.mark.asyncio
async def test_models_with_base_model(asyncpg_conn):
    from frankfurt import Database, fields
    from frankfurt.models import Model

    class BaseModel(Model):
        pk = fields.UUIDField(primary_key=True)

        class Meta:
            abstract = True

    class Writer(BaseModel):
        class Meta:
            table_name = 'writers'

    class Book(BaseModel):
        author = fields.ForeignKeyField(to='writers')

    db = await Database('__url__', models=[Writer, Book])

    # Create the tables.
    async with db.acquire() as conn:
        await conn.create_all_tables()

    assert asyncpg_conn.execute.mock_calls == [
        call('CREATE TABLE IF NOT EXISTS "writers" '
             '("pk" UUID PRIMARY KEY)'),
        call(f'CREATE TABLE IF NOT EXISTS "book" ('
             '"pk" UUID PRIMARY KEY, '
             '"author" UUID REFERENCES "writers" ("pk"))'),
    ]


@pytest.mark.asyncio
async def test_select_from_model(asyncpg_conn):
    from frankfurt import Database, fields
    from frankfurt.models import Model

    pk = uuid4()

    asyncpg_conn.fetch.return_value = [{'text': 'hola', 'pk': pk}]

    class Model(Model):
        pk = fields.UUIDField(primary_key=True)
        text = fields.CharField(max_length=200)

        class Meta:
            table_name = 'models'

    db = await Database('__url__', models=[Model])

    async with db.acquire() as conn:

        # Get one model.
        model = await conn.select(Model).where(text='hola').one()

    assert isinstance(model, Model), ".one should return the right instance."
    assert model['text'] == 'hola'
    assert 'pk' in model._data, "pk should be in _data"

    asyncpg_conn.fetch.assert_awaited_once_with(
        'SELECT "pk", "text" FROM "models" WHERE "text"=$1', 'hola'
    )


@pytest.mark.asyncio
async def test_insert_model(asyncpg_conn):

    from frankfurt import Database, fields
    from frankfurt.models import Model

    class Model(Model):
        pk = fields.UUIDField(primary_key=True)
        text = fields.CharField(max_length=200)

        class Meta:
            table_name = 'models'

    db = await Database('__url__', models=[Model])

    # Create a model.
    model = Model(text='hello!')

    # Set up the mock.
    asyncpg_conn.fetchrow.return_value = {
        'row': (model['pk'], 'hello!')
    }

    # Get a session.
    async with db.acquire() as conn:
        await conn.insert(model)

    asyncpg_conn.fetchrow.assert_awaited_once_with(
        'INSERT INTO "models" ("pk", "text") VALUES ($1, $2) '
        'RETURNING ("pk", "text")',
        model['pk'], "hello!"
    )


@pytest.mark.asyncio
async def test_update_model(asyncpg_conn):
    from frankfurt import Database, fields
    from frankfurt.models import Model, UpdateMixin

    pk = uuid4()

    asyncpg_conn.fetchrow.return_value = {
        'text1': 'hello!',
        'pk': pk,
        'text2': None
    }

    class Model(Model, UpdateMixin):
        pk = fields.UUIDField(primary_key=True)
        text1 = fields.CharField(max_length=200)
        text2 = fields.CharField(max_length=200)

        class Meta:
            table_name = 'models'

    db = await Database('__url__', models=[Model])

    async with await db.acquire() as conn:

        # Create a model.
        model = Model(text1='hello!', pk=pk)

        # Explicit call to update.
        await conn.update(model)

    asyncpg_conn.fetchrow.assert_awaited_once_with(
        'UPDATE "models" SET "text1"=$1, "text2"=$2 WHERE "pk"=$3 '
        'RETURNING ("pk", "text1", "text2")',
        "hello!", None, pk
    )


@pytest.mark.asyncio
async def test_delete_model_prevent_full_table_delete(asyncpg_conn):
    from frankfurt import Database, fields
    from frankfurt.models import Model, DeleteMixin

    asyncpg_conn.fetch.return_value = {
        'text': 'hello!'
    }

    class Model(Model, DeleteMixin):
        pk = fields.UUIDField(primary_key=True)
        text1 = fields.CharField(max_length=200)
        text2 = fields.CharField(max_length=200)

        class Meta:
            table_name = 'models'

    db = await Database('__url__', models=[Model])

    async with db.acquire() as conn:

        # Create a model
        model = Model(text1='hello!')

        # Save the model assumes is an update.
        with pytest.raises(Exception):
            await conn.delete(model)

    asyncpg_conn.fetch.assert_not_awaited()


async def _test_delete_model(asyncpg_conn):
    from frankfurt import Database, fields
    from frankfurt.models import Model, DeleteMixin

    pk = uuid4()

    asyncpg_conn.fetch.return_value = {
        'row': (pk, 'hello!', None)
    }

    class Model(Model, DeleteMixin):
        pk = fields.UUIDField(primary_key=True)
        text1 = fields.CharField(max_length=200)
        text2 = fields.CharField(max_length=200)

        class Meta:
            table_name = 'models'

    db = await Database('__url__', models=[Model])

    return db, Model(text1='hello!', pk=pk), pk


@pytest.mark.asyncio
async def test_delete_model(asyncpg_conn):

    db, instance, pk = await _test_delete_model(asyncpg_conn)

    # Save the model assumes is an update.
    async with db.acquire() as conn:
        await conn.delete(instance)

    asyncpg_conn.fetch.assert_awaited_once_with(
        'DELETE FROM "models" WHERE "pk"=$1 '
        'RETURNING ("pk", "text1", "text2")', pk
    )


@pytest.mark.asyncio
async def test_delete_mixin(asyncpg_conn):

    db, instance, pk = await _test_delete_model(asyncpg_conn)

    # Save the model assumes is an update.
    async with db.acquire() as conn:
        await instance.delete(conn)

    asyncpg_conn.fetch.assert_awaited_once_with(
        'DELETE FROM "models" WHERE "pk"=$1 '
        'RETURNING ("pk", "text1", "text2")', pk
    )


@pytest.mark.asyncio
async def test_create_mixin(asyncpg_conn):

    # Set tup the result from fetchrow.
    pk = uuid4()
    asyncpg_conn.fetchrow.return_value = {
        'row': (pk, )
    }

    from frankfurt import Database, fields
    from frankfurt.models import Model, CreateMixin

    class Book(Model, CreateMixin):
        pk = fields.UUIDField(primary_key=True)

    # Start the database.
    db = await Database('__url__', models=(Book,))

    # Acquire a connection/session.
    conn = await db.acquire()

    # Create a book.
    await Book.create(conn, pk=pk)

    # Assert the calls to db.
    asyncpg_conn.fetchrow.assert_awaited_once_with(
        'INSERT INTO "book" ("pk") VALUES ($1) '
        'RETURNING ("pk")',
        pk
    )
